@@ -34,24 +34,26 @@
 (declare-function agent-shell--interpolate-gradient "agent-shell")
 (declare-function agent-shell--make-acp-client "agent-shell")
 (declare-function agent-shell-make-agent-config "agent-shell")
-(declare-function agent-shell-start "agent-shell")
+(declare-function agent-shell--dwim "agent-shell")
 
-(cl-defun agent-shell-google-make-authentication (&key api-key login vertex-ai)
+(cl-defun agent-shell-google-make-authentication (&key api-key login vertex-ai none)
   "Create Google authentication configuration.
 
 API-KEY is the Google API key string or function that returns it.
 LOGIN when non-nil indicates to use login-based authentication.
 VERTEX-AI when non-nil indicates to use Vertex AI authentication.
+NONE when non-nil indicates no authentication method is used.
 
-Only one of API-KEY, LOGIN, or VERTEX-AI should be provided."
+Only one of API-KEY, LOGIN, VERTEX-AI, or NONE should be provided."
   (when (> (seq-count #'identity (list api-key login vertex-ai)) 1)
     (error "Cannot specify multiple authentication methods - choose one"))
-  (unless (> (seq-count #'identity (list api-key login vertex-ai)) 0)
+  (unless (> (seq-count #'identity (list api-key login vertex-ai none)) 0)
     (error "Must specify one of :api-key, :login, or :vertex-ai"))
   (cond
    (api-key `((:api-key . ,api-key)))
    (login `((:login . t)))
-   (vertex-ai `((:vertex-ai . t)))))
+   (vertex-ai `((:vertex-ai . t)))
+   (none `((:none . t)))))
 
 (defcustom agent-shell-google-authentication
   (agent-shell-google-make-authentication :login t)
@@ -75,7 +77,12 @@ For API key (function):
 For Vertex AI authentication:
 
   (setq agent-shell-google-authentication
-        (agent-shell-google-make-authentication :vertex-ai t))"
+        (agent-shell-google-make-authentication :vertex-ai t))
+
+For no authentication (when using alternative authentication methods):
+
+  (setq agent-shell-google-authentication
+        (agent-shell-google-make-authentication :none t))"
   :type 'alist
   :group 'agent-shell)
 
@@ -118,14 +125,37 @@ Returns an agent configuration alist using `agent-shell-make-agent-config'."
    :shell-prompt-regexp "Gemini> "
    :icon-name "gemini.png"
    :welcome-function #'agent-shell-google--gemini-welcome-message
-   :needs-authentication t
+   :needs-authentication (not (map-elt agent-shell-google-authentication :none))
    :authenticate-request-maker (lambda ()
                                  (cond ((map-elt agent-shell-google-authentication :api-key)
-                                        (acp-make-authenticate-request :method-id "gemini-api-key"))
+                                        ;; TODO: Save authentication methods from
+                                        ;; initialization and resolve :method-id
+                                        ;; to :method which came from the agent.
+                                        (acp-make-authenticate-request
+                                         :method-id "gemini-api-key"
+                                         :method '((id . "gemini-api-key")
+                                                   (name . "Use Gemini API key")
+                                                   (description . "Requires setting the `GEMINI_API_KEY` environment variable"))))
                                        ((map-elt agent-shell-google-authentication :vertex-ai)
-                                        (acp-make-authenticate-request :method-id "vertex-ai"))
+                                        ;; TODO: Save authentication methods from
+                                        ;; initialization and resolve :method-id
+                                        ;; to :method which came from the agent.
+                                        (acp-make-authenticate-request
+                                         :method-id "vertex-ai"
+                                         :method '((id . "vertex-ai")
+                                                   (name . "Vertex AI")
+                                                   (description . ""))))
+                                       ((map-elt agent-shell-google-authentication :none)
+                                        nil)
                                        (t
-                                        (acp-make-authenticate-request :method-id "oauth-personal"))))
+                                        ;; TODO: Save authentication methods from
+                                        ;; initialization and resolve :method-id
+                                        ;; to :method which came from the agent.
+                                        (acp-make-authenticate-request
+                                         :method-id "oauth-personal"
+                                         :method '((id . "oauth-personal")
+                                                   (name . "Log in with Google")
+                                                   (description . ""))))))
    :client-maker (lambda (buffer)
                    (agent-shell-google-make-gemini-client :buffer buffer))
    :install-instructions "See https://github.com/google-gemini/gemini-cli for installation."))
@@ -133,8 +163,8 @@ Returns an agent configuration alist using `agent-shell-make-agent-config'."
 (defun agent-shell-google-start-gemini ()
   "Start an interactive Gemini CLI agent shell."
   (interactive)
-  (agent-shell-start
-   :config (agent-shell-google-make-gemini-config)))
+  (agent-shell--dwim :config (agent-shell-google-make-gemini-config)
+                     :new-shell t))
 
 (cl-defun agent-shell-google-make-gemini-client (&key buffer)
   "Create a Gemini client using configured authentication with BUFFER as context.
@@ -162,6 +192,11 @@ Uses `agent-shell-google-authentication' for authentication configuration."
                                   :command-params (cdr agent-shell-google-gemini-command)
                                   :environment-variables agent-shell-google-gemini-environment
                                   :context-buffer buffer))
+   ((map-elt agent-shell-google-authentication :none)
+    (agent-shell--make-acp-client :command (car agent-shell-google-gemini-command)
+                                  :command-params (cdr agent-shell-google-gemini-command)
+                                  :environment-variables agent-shell-google-gemini-environment
+                                  :context-buffer buffer))
    (t
     (error "Invalid authentication configuration"))))
 
@@ -204,7 +239,7 @@ Uses `agent-shell-google-authentication' for authentication configuration."
                  (color (agent-shell--interpolate-gradient gradient-colors progress)))
             (setq propertized-line
                   (concat propertized-line
-                          (propertize char 'font-lock-face `(:foreground ,color))))))
+                          (propertize char 'font-lock-face `(:foreground ,color :inherit fixed-pitch))))))
         (setq result (concat result propertized-line "\n"))))
     (string-trim-right result)))
 
@@ -219,7 +254,7 @@ Uses `agent-shell-google-authentication' for authentication configuration."
     (dotimes (i (length text))
       (setq result (concat result
                            (propertize (substring text i (1+ i))
-                                       'font-lock-face `(:foreground ,(nth (mod i (length colors)) colors))))))
+                                       'font-lock-face `(:foreground ,(nth (mod i (length colors)) colors) :inherit fixed-pitch)))))
     result))
 
 (defun agent-shell-google-key ()
