@@ -44,7 +44,7 @@
 (declare-function agent-shell--get-region "agent-shell")
 (declare-function agent-shell--insert-to-shell-buffer "agent-shell")
 (declare-function agent-shell--make-header "agent-shell")
-(declare-function agent-shell--context "agent-shell")
+(declare-function agent-shell--relevant-text "agent-shell")
 (declare-function agent-shell--shell-buffer "agent-shell")
 (declare-function agent-shell--start "agent-shell")
 (declare-function agent-shell--state "agent-shell")
@@ -61,27 +61,15 @@
 (declare-function agent-shell-ui-backward-block "agent-shell")
 (declare-function agent-shell-ui-forward-block "agent-shell")
 (declare-function agent-shell-ui-mode "agent-shell")
-(declare-function agent-shell-completion-mode "agent-shell-completion")
-(declare-function agent-shell-yank-dwim "agent-shell")
 
 (defvar agent-shell-header-style)
 (defvar agent-shell-prefer-viewport-interaction)
 (defvar agent-shell-preferred-agent-config)
-(defvar agent-shell-session-strategy)
-(defvar agent-shell--state)
-(defvar agent-shell-file-completion-enabled)
 
-(defvar-local agent-shell-viewport--compose-snapshot nil
-  "Alist with :content and :location from compose buffer before viewing history.")
-;; The viewport buffer transitions between major modes which clears
-;; buffer-local vars. Make snapshot permanent-local.
-(put 'agent-shell-viewport--compose-snapshot 'permanent-local t)
-
-(cl-defun agent-shell-viewport--show-buffer (&key append override submit no-focus shell-buffer)
+(cl-defun agent-shell-viewport--show-buffer (&key text submit no-focus shell-buffer)
   "Show a viewport compose buffer for the agent shell.
 
-APPEND is appended to the viewport compose buffer.
-OVERRIDE, when non-nil, replaces content verbatim (no trimming).
+TEXT is inserted into the viewport compose buffer.
 SUBMIT, when non-nil, submits after insertion.
 NO-FOCUS, when non-nil, avoids focusing the viewport compose buffer.
 SHELL-BUFFER, when non-nil, prefer this shell buffer.
@@ -96,73 +84,44 @@ Returns an alist with insertion details or nil otherwise:
     (error "Not yet supported"))
   (when no-focus
     (error "Not yet supported"))
-  (when (and append override)
-    (error "Use :append or :override but not both"))
   (when shell-buffer
     ;; Momentarily set buffer to same window, so it's recent in stack.
     (let ((current (current-buffer)))
       (pop-to-buffer-same-window shell-buffer)
       (pop-to-buffer-same-window current)))
-  (when-let* ((shell-buffer (or shell-buffer (agent-shell--shell-buffer)))
-              (viewport-buffer (agent-shell-viewport--buffer :shell-buffer shell-buffer))
-              (text (or append (agent-shell--context :shell-buffer shell-buffer) "")))
-    (when (and override (not (string-empty-p text)))
-      (error "Cannot override"))
+  (when-let ((shell-buffer (or shell-buffer (agent-shell--shell-buffer)))
+             (viewport-buffer (agent-shell-viewport--buffer :shell-buffer shell-buffer))
+             (text (or text (agent-shell--relevant-text) "")))
     (let ((insert-start nil)
           (insert-end nil))
       ;; Is there text to be inserted? Reject while busy.
       (when (and (agent-shell-viewport--busy-p
                   :viewport-buffer viewport-buffer)
-                 (or (not (string-empty-p (string-trim text)))
-                     (and override (not (string-empty-p (string-trim override))))))
+                 (not (string-empty-p (string-trim text))))
         (user-error "Busy... please wait"))
       (agent-shell--display-buffer viewport-buffer)
-      (when (and override
-                 (with-current-buffer viewport-buffer
-                   ;; viewport buffer empty?
-                   (not (= (buffer-size) 0))))
-        (unless (y-or-n-p "Compose buffer is not empty.  Override?")
-          ;; User does not want to override.
-          ;; Treat as regurlar text (typically appended).
-          (setq text (concat text
-                             (unless (string-empty-p text)
-                               "\n\n")
-                             override))
-          (setq override nil)))
       ;; TODO: Do we need to get prompt and partial response,
       ;; in case viewport compose buffer is created for the
       ;; first time on an ongoing/busy shell session?
-      (cond
-       ((agent-shell-viewport--busy-p)
-        (agent-shell-viewport-view-mode))
-       (override
-        (agent-shell-viewport-edit-mode)
-        (agent-shell-viewport--initialize)
-        (setq insert-start (point))
-        (insert override)
-        (setq insert-end (point)))
-       ((derived-mode-p 'agent-shell-viewport-edit-mode)
-        (unless (string-empty-p text)
+      (if (agent-shell-viewport--busy-p)
+          (agent-shell-viewport-view-mode)
+        (if (derived-mode-p 'agent-shell-viewport-edit-mode)
+            (unless (string-empty-p text)
+              (save-excursion
+                (goto-char (point-max))
+                (setq insert-start (point))
+                (unless (string-empty-p text)
+                  (insert "\n\n" text))
+                (setq insert-end (point))))
+          (agent-shell-viewport-edit-mode)
+          ;; Transitioned to edit mode. Start empty.
+          (agent-shell-viewport--initialize)
           (save-excursion
             (goto-char (point-max))
             (setq insert-start (point))
-            (insert "\n\n" text)
+            (unless (string-empty-p text)
+              (insert "\n\n" text))
             (setq insert-end (point)))))
-       (t
-        (agent-shell-viewport-edit-mode)
-        ;; Transitioned to edit mode. Wipe content.
-        (agent-shell-viewport--initialize)
-        ;; Restore snapshot if needed.
-        (when-let ((snapshot agent-shell-viewport--compose-snapshot))
-          (insert (map-elt snapshot :content))
-          (goto-char (map-elt snapshot :location))
-          (setq agent-shell-viewport--compose-snapshot nil))
-        (save-excursion
-          (goto-char (point-max))
-          (setq insert-start (point))
-          (unless (string-empty-p text)
-            (insert "\n\n" text))
-          (setq insert-end (point)))))
       `((:buffer . ,viewport-buffer)
         (:start . ,insert-start)
         (:end . ,insert-end)))))
@@ -170,13 +129,6 @@ Returns an alist with insertion details or nil otherwise:
 (defun agent-shell-viewport-compose-send ()
   "Send the viewport composed prompt to the agent shell."
   (interactive)
-  (unless (derived-mode-p 'agent-shell-viewport-edit-mode)
-    (user-error "Not in a shell viewport buffer"))
-  (when (and (not (eq agent-shell-session-strategy 'new-deferred))
-             (not (with-current-buffer (agent-shell-viewport--shell-buffer)
-                    (map-nested-elt agent-shell--state '(:session :id)))))
-    (user-error "Session not ready... please wait"))
-  (setq agent-shell-viewport--compose-snapshot nil)
   (if agent-shell-prefer-viewport-interaction
       (agent-shell-viewport-compose-send-and-wait-for-response)
     (agent-shell-viewport-compose-send-and-kill)))
@@ -184,9 +136,7 @@ Returns an alist with insertion details or nil otherwise:
 (defun agent-shell-viewport-compose-send-and-kill ()
   "Send the viewport composed prompt to the agent shell and kill compose buffer."
   (interactive)
-  (unless (derived-mode-p 'agent-shell-viewport-edit-mode)
-    (user-error "Not in a shell viewport buffer"))
-  (let ((shell-buffer (agent-shell-viewport--shell-buffer))
+  (let ((shell-buffer (agent-shell--shell-buffer))
         (viewport-buffer (current-buffer))
         (prompt (buffer-string)))
     (with-current-buffer shell-buffer
@@ -202,8 +152,8 @@ Returns an alist with insertion details or nil otherwise:
   (catch 'exit
     (unless (derived-mode-p 'agent-shell-viewport-edit-mode)
       (user-error "Not in a shell viewport buffer"))
-    (let ((shell-buffer (agent-shell-viewport--shell-buffer))
-          (viewport-buffer (current-buffer))
+    (let ((shell-buffer (agent-shell--shell-buffer))
+          (viewport-buffer (agent-shell-viewport--buffer))
           (prompt (string-trim (buffer-string))))
       (when (agent-shell-viewport--busy-p)
         (unless (y-or-n-p "Interrupt?")
@@ -231,7 +181,6 @@ Returns an alist with insertion details or nil otherwise:
           (agent-shell-viewport-edit-mode)
           (agent-shell-viewport--initialize))
         (agent-shell--insert-to-shell-buffer
-         :shell-buffer (agent-shell-viewport--shell-buffer)
          :text prompt
          :submit t
          :no-focus t)
@@ -242,9 +191,8 @@ Returns an alist with insertion details or nil otherwise:
 (defun agent-shell-viewport-interrupt ()
   "Interrupt active agent shell request."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
   (catch 'exit
-    (let ((shell-buffer (agent-shell-viewport--shell-buffer)))
+    (let ((shell-buffer (agent-shell--shell-buffer)))
       (unless (agent-shell-viewport--busy-p)
         (user-error "No pending request"))
       (unless (y-or-n-p "Interrupt?")
@@ -257,14 +205,15 @@ Returns an alist with insertion details or nil otherwise:
   "Initialize viewport compose buffer.
 
 Optionally set its PROMPT and RESPONSE."
-  (agent-shell-viewport--ensure-buffer)
-
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
   ;; Recalculate and cache position
   (agent-shell-viewport--position :force-refresh t)
   (let ((inhibit-read-only t)
         (viewport-buffer (current-buffer)))
     (erase-buffer)
-    (when-let ((shell-buffer (agent-shell-viewport--shell-buffer)))
+    (when-let ((shell-buffer (agent-shell--shell-buffer)))
       (with-current-buffer shell-buffer
         (unless (eq agent-shell-header-style 'graphical)
           ;; Insert read-only newline at the point-min
@@ -289,12 +238,6 @@ Optionally set its PROMPT and RESPONSE."
     (let ((inhibit-read-only t))
       (markdown-overlays-put))))
 
-(defun agent-shell-viewport--ensure-buffer ()
-  "Ensure current buffer is a viewport and err otherwise."
-  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
-              (derived-mode-p 'agent-shell-viewport-edit-mode))
-    (user-error "Not in a shell viewport buffer")))
-
 (defun agent-shell-viewport--prompt ()
   "Return the buffer prompt."
   (save-excursion
@@ -346,41 +289,25 @@ Optionally set its PROMPT and RESPONSE."
 (defun agent-shell-viewport-compose-cancel ()
   "Cancel prompt composition."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
-  (setq agent-shell-viewport--compose-snapshot nil)
-  (let ((viewport-buffer (current-buffer))
-        (shell-buffer (agent-shell-viewport--shell-buffer)))
-    ;; View mode
-    (if (or (derived-mode-p 'agent-shell-viewport-view-mode)
-            (with-current-buffer shell-buffer
-              (not (shell-maker-history))))
-        (bury-buffer)
-      ;; Edit mode
-      (when (or (string-empty-p (string-trim (buffer-string)))
-                (y-or-n-p "Discard composed prompt? "))
-        (if agent-shell-prefer-viewport-interaction
-            (agent-shell-viewport-view-last)
-          (agent-shell-other-buffer)
-          (kill-buffer viewport-buffer))))))
-
-(defun agent-shell-viewport-compose-peek-last ()
-  "Save compose buffer snapshot and peek at the last interaction."
-  (interactive)
-  (unless (derived-mode-p 'agent-shell-viewport-edit-mode)
-    (user-error "Not in a prompt compose buffer"))
-  (unless (with-current-buffer (agent-shell-viewport--shell-buffer)
-            (shell-maker-history))
-    (user-error "No items in history"))
-  (setq agent-shell-viewport--compose-snapshot
-        `((:content . ,(buffer-string))
-          (:location . ,(point))))
-  (agent-shell-viewport-view-last))
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
+  (if (or (derived-mode-p 'agent-shell-viewport-view-mode)
+          (with-current-buffer (agent-shell--shell-buffer)
+            (not (shell-maker-history))))
+      (bury-buffer)
+    ;; Edit mode
+    (when (or (string-empty-p (string-trim (buffer-string)))
+              (y-or-n-p "Discard composed prompt? "))
+      (agent-shell-viewport-view-last))))
 
 (defun agent-shell-viewport-view-last ()
   "Display the last request/response interaction."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
-  (when-let ((shell-buffer (agent-shell-viewport--shell-buffer)))
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
+  (when-let ((shell-buffer (agent-shell--shell-buffer)))
     (with-current-buffer shell-buffer
       (goto-char comint-last-input-start)))
   (agent-shell-viewport-view-mode)
@@ -389,9 +316,11 @@ Optionally set its PROMPT and RESPONSE."
 (defun agent-shell-viewport-refresh ()
   "Refresh viewport buffer content with current item from shell."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
-  (when-let ((shell-buffer (agent-shell-viewport--shell-buffer))
-             (viewport-buffer (current-buffer))
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
+  (when-let ((shell-buffer (agent-shell--shell-buffer))
+             (viewport-buffer (agent-shell-viewport--buffer))
              (current (with-current-buffer shell-buffer
                         (or (shell-maker--command-and-response-at-point)
                             (shell-maker-next-command-and-response t)))))
@@ -480,9 +409,6 @@ If at the first item, attempt to switch to previous interaction."
          (when prompt-start
            (goto-char prompt-start)))))))
 
-(defconst agent-shell-viewport--suffix " [viewport]"
-  "Suffix appended to shell buffer name to create viewport buffer name.")
-
 (cl-defun agent-shell-viewport--buffer (&key shell-buffer existing-only)
   "Get the viewport buffer associated with a SHELL-BUFFER.
 
@@ -490,8 +416,8 @@ With EXISTING-ONLY, only return existing buffers without creating."
   (when-let ((shell-buffer (or shell-buffer
                                (agent-shell--shell-buffer))))
     (with-current-buffer shell-buffer
-      (let* ((viewport-buffer-name (concat (buffer-name (get-buffer shell-buffer))
-                                           agent-shell-viewport--suffix))
+      (let* ((viewport-buffer-name (concat (buffer-name shell-buffer)
+                                          " [viewport]"))
              (viewport-buffer (get-buffer viewport-buffer-name)))
         (if viewport-buffer
             viewport-buffer
@@ -504,8 +430,6 @@ With EXISTING-ONLY, only return existing buffers without creating."
 (defun agent-shell-viewport-reply ()
   "Reply as a follow-up and compose another prompt/query."
   (interactive)
-  (unless (derived-mode-p 'agent-shell-viewport-view-mode)
-    (user-error "Not in a shell viewport buffer"))
   (when (agent-shell-viewport--busy-p)
     (user-error "Busy, please wait"))
   (let* ((region (map-elt (agent-shell--get-region :deactivate t) :content))
@@ -516,37 +440,18 @@ With EXISTING-ONLY, only return existing buffers without creating."
                                           (split-string region "\n")
                                           "\n")
                                "\n\n"))))
-    (with-current-buffer (agent-shell-viewport--shell-buffer)
+    (with-current-buffer (agent-shell--shell-buffer)
       (goto-char (point-max)))
-    (let ((snapshot agent-shell-viewport--compose-snapshot))
-      (agent-shell-viewport-edit-mode)
+    (agent-shell-viewport-edit-mode)
+    (if block-quoted-text
+        (progn
+          (agent-shell-viewport--initialize :prompt block-quoted-text)
+          (goto-char (point-max)))
       (agent-shell-viewport--initialize)
-      (when snapshot
-        (insert (map-elt snapshot :content))
-        (setq agent-shell-viewport--compose-snapshot nil))
-      (when block-quoted-text
-        (goto-char (point-max))
-        (insert (if snapshot "\n\n" "") block-quoted-text))
-      (goto-char (if (or snapshot block-quoted-text)
-                     (point-max)
-                   (point-min))))
+      (goto-char (point-min)))
     ;; Setting point isn't enough at times. Force scrolling.
     (set-window-start (selected-window) (point-min))))
 
-(defun agent-shell-viewport-reply-yes ()
-  "Reply with \"yes\" and send immediately."
-  (interactive)
-  (agent-shell-viewport-reply)
-  (insert "yes")
-  (agent-shell-viewport-compose-send))
-
-(defun agent-shell-viewport-reply-more ()
-  "Reply with \"more\" and send immediately."
-  (interactive)
-  (agent-shell-viewport-reply)
-  (insert "more")
-  (agent-shell-viewport-compose-send))
-
 (defun agent-shell-viewport-previous-page ()
   "Show previous interaction (request / response)."
   (interactive)
@@ -556,53 +461,38 @@ With EXISTING-ONLY, only return existing buffers without creating."
   "Show next interaction (request / response).
 
 If BACKWARDS is non-nil, go to previous interaction.
-If START-AT-TOP is non-nil, position at point-min regardless of direction.
-
-If there are no more next items and a compose snapshot exists, restore the
-buffer from the snapshot and switch to edit mode."
+If START-AT-TOP is non-nil, position at point-min regardless of direction."
   (interactive)
   (unless (derived-mode-p 'agent-shell-viewport-view-mode)
     (error "Not in a viewport buffer"))
   (when (agent-shell-viewport--busy-p)
     (user-error "Busy... please wait"))
-  (let ((shell-buffer (agent-shell-viewport--shell-buffer))
-        (snapshot agent-shell-viewport--compose-snapshot)
-        (pos (agent-shell-viewport--position :force-refresh t)))
-    ;; Check if at last position going forward with a snapshot to restore
-    (if (and (not backwards) snapshot pos
-             (= (car pos) (cdr pos)))
-        (progn
-          (agent-shell-viewport-edit-mode)
-          (agent-shell-viewport--initialize)
-          (insert (map-elt snapshot :content))
-          (goto-char (map-elt snapshot :location))
-          (setq agent-shell-viewport--compose-snapshot nil)
-          (cl-return-from agent-shell-viewport-next-page))
-      (when-let ((next (with-current-buffer shell-buffer
-                         (if backwards
-                             (when (save-excursion
-                                     (let ((orig-line (line-number-at-pos)))
-                                       (comint-previous-prompt 1)
-                                       (= orig-line (line-number-at-pos))))
-                               (error "No previous page"))
-                           (when (save-excursion
-                                   (let ((orig-line (point)))
-                                     (comint-next-prompt 1)
-                                     (= orig-line (point))))
-                             (error "No next page")))
-                         (shell-maker-next-command-and-response backwards))))
-        (agent-shell-viewport--initialize
-         :prompt (car next) :response (cdr next))
-        (goto-char (if start-at-top
-                       (point-min)
-                     (if backwards (point-max) (point-min))))
-        (agent-shell-viewport--update-header)
-        next))))
+  (when-let ((shell-buffer (agent-shell--shell-buffer))
+             (viewport-buffer (agent-shell-viewport--buffer))
+             (next (with-current-buffer shell-buffer
+                     (if backwards
+                         (when (save-excursion
+                                 (let ((orig-line (line-number-at-pos)))
+                                   (comint-previous-prompt 1)
+                                   (= orig-line (line-number-at-pos))))
+                           (error "No previous page"))
+                       (when (save-excursion
+                               (let ((orig-line (point)))
+                                 (comint-next-prompt 1)
+                                 (= orig-line (point))))
+                         (error "No next page")))
+                     (shell-maker-next-command-and-response backwards))))
+    (agent-shell-viewport--initialize
+     :prompt (car next) :response (cdr next))
+    (goto-char (if start-at-top
+                   (point-min)
+                 (if backwards (point-max) (point-min))))
+    (agent-shell-viewport--update-header)
+    next))
 
 (defun agent-shell-viewport-set-session-model ()
   "Set session model."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
   (let* ((shell-buffer (or (agent-shell--current-shell)
                            (user-error "Not in an agent-shell buffer")))
          (viewport-buffer (agent-shell-viewport--buffer
@@ -617,7 +507,6 @@ buffer from the snapshot and switch to edit mode."
 (defun agent-shell-viewport-set-session-mode ()
   "Set session mode."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
   (let* ((shell-buffer (or (agent-shell--current-shell)
                            (user-error "Not in an agent-shell buffer")))
          (viewport-buffer (agent-shell-viewport--buffer
@@ -633,7 +522,6 @@ buffer from the snapshot and switch to edit mode."
 (defun agent-shell-viewport-cycle-session-mode ()
   "Cycle through available session modes."
   (interactive)
-  (agent-shell-viewport--ensure-buffer)
   (let* ((shell-buffer (or (agent-shell--current-shell)
                            (user-error "Not in an agent-shell buffer")))
          (viewport-buffer (agent-shell-viewport--buffer
@@ -646,18 +534,16 @@ buffer from the snapshot and switch to edit mode."
            (with-current-buffer viewport-buffer
              (agent-shell-viewport--update-header))))))))
 
-;; Continuously fetching position can get expensive. Cache it.
-(defvar-local agent-shell-viewport--position-cache nil
-  "Cached position value (CURRENT . TOTAL).")
-
 (cl-defun agent-shell-viewport--position (&key force-refresh)
   "Return the position in history of the shell buffer.
 
 When FORCE-REFRESH is non-nil, recalculate and update cache."
-  (agent-shell-viewport--ensure-buffer)
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
   (if (and (not force-refresh) agent-shell-viewport--position-cache)
       agent-shell-viewport--position-cache
-    (let* ((shell-buffer (agent-shell-viewport--shell-buffer))
+    (let* ((shell-buffer (agent-shell--shell-buffer))
            (current (with-current-buffer shell-buffer
                       (shell-maker--command-and-response-at-point)))
            (history (with-current-buffer shell-buffer
@@ -681,45 +567,13 @@ VIEWPORT-BUFFER is the viewport buffer to check."
     (with-current-buffer shell-buffer
       shell-maker--busy)))
 
-(defvar agent-shell-viewport-edit-mode-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map (kbd "C-c C-c") #'agent-shell-viewport-compose-send)
-    (define-key map (kbd "C-c C-p") #'agent-shell-viewport-compose-peek-last)
-    (define-key map (kbd "C-c C-k") #'agent-shell-viewport-compose-cancel)
-    (define-key map (kbd "C-<tab>") #'agent-shell-viewport-cycle-session-mode)
-    (define-key map (kbd "C-c C-m") #'agent-shell-viewport-set-session-mode)
-    (define-key map (kbd "C-c C-v") #'agent-shell-viewport-set-session-model)
-    (define-key map (kbd "C-c C-o") #'agent-shell-other-buffer)
-    (define-key map [remap yank] #'agent-shell-yank-dwim)
-    map)
-  "Keymap for `agent-shell-viewport-edit-mode'.")
-
-(defvar agent-shell-viewport-view-mode-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map (kbd "C-c C-c") #'agent-shell-viewport-interrupt)
-    (define-key map (kbd "TAB") #'agent-shell-viewport-next-item)
-    (define-key map (kbd "<backtab>") #'agent-shell-viewport-previous-item)
-    (define-key map (kbd "n") #'agent-shell-viewport-next-item)
-    (define-key map (kbd "p") #'agent-shell-viewport-previous-item)
-    (define-key map (kbd "f") #'agent-shell-viewport-next-page)
-    (define-key map (kbd "b") #'agent-shell-viewport-previous-page)
-    (define-key map (kbd "r") #'agent-shell-viewport-reply)
-    (define-key map (kbd "y") #'agent-shell-viewport-reply-yes)
-    (define-key map (kbd "q") #'bury-buffer)
-    (define-key map (kbd "C-<tab>") #'agent-shell-viewport-cycle-session-mode)
-    (define-key map (kbd "v") #'agent-shell-viewport-set-session-model)
-    (define-key map (kbd "m") #'agent-shell-viewport-reply-more)
-    (define-key map (kbd "s") #'agent-shell-viewport-set-session-mode)
-    (define-key map (kbd "o") #'agent-shell-other-buffer)
-    (define-key map (kbd "C-c C-o") #'agent-shell-other-buffer)
-    map)
-  "Keymap for `agent-shell-viewport-view-mode'.")
-
 (defun agent-shell-viewport--update-header ()
   "Update header and mode line based on `agent-shell-header-style'.
 
 Automatically determines qualifier and bindings based on current major mode."
-  (agent-shell-viewport--ensure-buffer)
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
   (let* ((pos (or (agent-shell-viewport--position)
                   (cons 1 1)))
          (pos-label (format "%d/%d" (car pos) (cdr pos)))
@@ -740,11 +594,7 @@ Automatically determines qualifier and bindings based on current major mode."
                       `((:key . ,(key-description (where-is-internal
                                                    'agent-shell-viewport-compose-cancel
                                                    agent-shell-viewport-edit-mode-map t)))
-                        (:description . "cancel"))
-                      `((:key . ,(key-description (where-is-internal
-                                                   'agent-shell-viewport-compose-peek-last
-                                                   agent-shell-viewport-edit-mode-map t)))
-                        (:description . "previous page"))))
+                        (:description . "cancel"))))
                     ((derived-mode-p 'agent-shell-viewport-view-mode)
                      (append
                       (list
@@ -768,7 +618,7 @@ Automatically determines qualifier and bindings based on current major mode."
                                                       'agent-shell-viewport-interrupt
                                                       agent-shell-viewport-view-mode-map t)))
                            (:description . "interrupt")))))))))
-    (when-let* ((shell-buffer (agent-shell-viewport--shell-buffer))
+    (when-let* ((shell-buffer (agent-shell--shell-buffer))
                 (header (with-current-buffer shell-buffer
                           (cond
                            ((eq agent-shell-header-style 'graphical)
@@ -783,23 +633,17 @@ Automatically determines qualifier and bindings based on current major mode."
 
 (defvar-local agent-shell-viewport--clean-up t)
 
-(cl-defun agent-shell-viewport--shell-buffer (&optional viewport-buffer)
-  "Get the shell buffer associated with VIEWPORT-BUFFER.
-
-Derives shell buffer name by removing the viewport suffix from buffer name.
-Returns nil if VIEWPORT-BUFFER is not a viewport buffer or shell doesn't exist."
-  (when-let* ((viewport-name (buffer-name (or viewport-buffer (current-buffer))))
-              ((string-suffix-p agent-shell-viewport--suffix viewport-name))
-              (shell-name (substring viewport-name 0
-                                     (- (length viewport-name)
-                                        (length agent-shell-viewport--suffix)))))
-    (get-buffer shell-name)))
+;; Continuously fetching position can get expensive. Cache it.
+(defvar-local agent-shell-viewport--position-cache nil
+  "Cached position value (CURRENT . TOTAL).")
 
 (defun agent-shell-viewport--clean-up ()
   "Clean up resources.
 
 For example, offer to kill associated shell session."
-  (agent-shell-viewport--ensure-buffer)
+  (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
+              (derived-mode-p 'agent-shell-viewport-edit-mode))
+    (user-error "Not in a shell viewport buffer"))
   (if (and agent-shell-viewport--clean-up
            ;; Only offer to kill shell buffers when viewport buffer
            ;; is explicitly being killed from a viewport buffer.
@@ -809,33 +653,57 @@ For example, offer to kill associated shell session."
       ;; triggered by shell buffers attempting to kill viewport buffer.
       (let ((agent-shell-viewport--clean-up nil))
         (when-let ((shell-buffers (seq-filter (lambda (shell-buffer)
-                                                (and (equal (agent-shell-viewport--buffer
-                                                             :shell-buffer shell-buffer
-                                                             :existing-only t)
-                                                            (current-buffer))
-                                                     ;; Skip shells already shutting down (client
-                                                     ;; is nil after agent-shell--shutdown).
-                                                     (buffer-local-value 'agent-shell--state shell-buffer)
-                                                     (map-elt (buffer-local-value 'agent-shell--state shell-buffer) :client)))
+                                                (equal (agent-shell-viewport--buffer
+                                                        :shell-buffer shell-buffer
+                                                        :existing-only t)
+                                                       (current-buffer)))
                                               (agent-shell-buffers)))
                    ((y-or-n-p "Kill shell session too?")))
           (mapc (lambda (shell-buffer)
                   (kill-buffer shell-buffer))
                 shell-buffers)))))
 
+(defvar agent-shell-viewport-edit-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "C-c C-c") #'agent-shell-viewport-compose-send)
+    (define-key map (kbd "C-c C-k") #'agent-shell-viewport-compose-cancel)
+    (define-key map (kbd "C-<tab>") #'agent-shell-viewport-cycle-session-mode)
+    (define-key map (kbd "C-c C-m") #'agent-shell-viewport-set-session-mode)
+    (define-key map (kbd "C-c C-v") #'agent-shell-viewport-set-session-model)
+    (define-key map (kbd "C-c C-o") #'agent-shell-other-buffer)
+    map)
+  "Keymap for `agent-shell-viewport-edit-mode'.")
+
 (define-derived-mode agent-shell-viewport-edit-mode text-mode "Agent Shell Viewport (Edit)"
   "Major mode for composing agent shell prompts.
 
 \\{agent-shell-viewport-edit-mode-map}"
   (cursor-intangible-mode +1)
   (setq buffer-read-only nil)
-  (when agent-shell-file-completion-enabled
-    (agent-shell-completion-mode +1))
   (agent-shell-viewport--update-header)
   (let ((inhibit-read-only t))
     (erase-buffer))
   (add-hook 'kill-buffer-hook #'agent-shell-viewport--clean-up nil t))
 
+(defvar agent-shell-viewport-view-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "C-c C-c") #'agent-shell-viewport-interrupt)
+    (define-key map (kbd "n") #'agent-shell-viewport-next-item)
+    (define-key map (kbd "p") #'agent-shell-viewport-previous-item)
+    (define-key map (kbd "<tab>") #'agent-shell-viewport-next-item)
+    (define-key map (kbd "<backtab>") #'agent-shell-viewport-previous-item)
+    (define-key map (kbd "f") #'agent-shell-viewport-next-page)
+    (define-key map (kbd "b") #'agent-shell-viewport-previous-page)
+    (define-key map (kbd "r") #'agent-shell-viewport-reply)
+    (define-key map (kbd "q") #'bury-buffer)
+    (define-key map (kbd "C-<tab>") #'agent-shell-viewport-cycle-session-mode)
+    (define-key map (kbd "v") #'agent-shell-viewport-set-session-model)
+    (define-key map (kbd "m") #'agent-shell-viewport-set-session-mode)
+    (define-key map (kbd "o") #'agent-shell-other-buffer)
+    (define-key map (kbd "C-c C-o") #'agent-shell-other-buffer)
+    map)
+  "Keymap for `agent-shell-viewport-view-mode'.")
+
 (define-derived-mode agent-shell-viewport-view-mode text-mode "Agent Shell Viewport (View)"
   "Major mode for viewing agent shell prompts (read-only).
 
